﻿#*********************************************************************
#*                                                                   *
#*   FTP website hack detection script                               *
#*   =================================                               *
#*                                                                   *
#*   psftpmon.ps1                                                    *
#*                                                                   *
#*                Jim Champ (c) 2016                                 *
#*  This is my first attempt at writing in powershell                *
#*  and I expect it shows. Improvements,suggestions welcomed         *
#*                                                                   *
#*            http://www.champwilde.f9.co.uk/idmphp/psftpmon         *
#*                                                                   *
#*                                                                   *
#*********************************************************************

<#
	University of Illinois/NCSA Open Source License for psftpmon.ps1

Copyright (c) 2016 Jim Champ
All rights reserved.
Developed by: 		Jim Champ
                    www.champwilde.f9.co.uk/idmphp/psftpmon/
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal with the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimers. 

Redistributions in binary form must reproduce the above
copyright notice, this list of conditions and the following disclaimers
in the documentation and/or other materials provided with the
distribution. 

The names of its contributors may not be used to endorse
or promote products derived from this Software without specific prior
written permission. 
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE. 

#>
 
 # script intended to monitor a web site via FTP and alert on changes
 # clearly not suitable for a busy dynamic site, intended to monitor a static site that isn't changing regularly
 # script maintains a local log file and compares no of files on server and last updated time
 # alerts when either change
 # also optionally notifies every n days to ensure its running
 # intended to be running as a background process on a windows client somewhere, probably with quite high frequency so intended low impact

 # ftp ideas from http://pleasework.robbievance.net/howto-get-timestamp-of-file-from-ftp-site-using-powershell/
 # and https://gist.github.com/arthurdent/0fd3880bd07e484a0af6
 # other ideas and lessons from various MS technet cotributors and stackoverflow. 
 # Thank you everyone

# Setting continue stops getting variable overwite error messages in IDE
$ErrorActionPreference = “SilentlyContinue”

# lets make the IDE tidy!
cls

#*********************************************************************
#*          CHANGE SETTINGS HERE                                     *
#*          ====================                                     *
#*                                                                   *
#*   Here we set constants, server authentication settings           *
#*      and everything else the normal user will need to change      *
#*      to suit their system and make the code work.                 *
#*                                                                   *
#*                                                                   *
#*********************************************************************

# this flag controls output to screen of major status messages. 
# should be $true or $false. Note the $. 
$displayflag = $true

# Here are the server details we are monitoring
# in an ideal world this user should have read only access to the web server
# since the password is a) hard coded and b) goes across the wire in clear text
# if you are using external hosting this may not be an option
$Server = "ftp.myisp.co.uk"
$User = "ftp"
$Pass = "ftppassword"
$BaseDirectory = "/srv/www/htdocs"
# important there's no trailing / on the foldername...
# This is unix syntax with / not \ as we are hitting an FTP system

#this is the interval of the confidence report that the system is still running
#and also the cycle on which the backup log file are rotated
#Valid intervals are "Daily", "Weekly", "Monthly", "Annual"
#note that confrdence/log changes are not truly scheduled. They run the first time the script is run after the digit changes, so will nornally be
#1st of Month, 1st January, after midnight etc...
$ConfidenceEmailInterval = "Monthly"
#this is the log location, which is Windows format as this is on the client
#make sure there is at least one \ in the path and a . in the file name or log rotation will fail
$LogFileLocation = "C:\logs\DonAshton.log"
#this is the number of backup log files to keep
$NoBackupLogs = 5

#These are smtp email addresses used by the alerts. Multiple addresses should be comma delimited
$DestUsers = "Tommy Atkins <tatkins@myisp.co.uk>"
$SrcUser = "PSFTPMon <donotreply@myisp.co.uk>"

# this is the smtp server for the mail alerts. The script assumes no authentication needed. If you need to 
# authenticate against smtp then it gets a fair bit more complicated. Don't much like having the ftp password in the script, but mail server 
# is even more disturbing 

$smtpserver = "relayserver.myisp.co.uk"



 function ExtractTimeStamps($TheTimeStamp) {
#*********************************************************************
#*                                                                   *
#*   This function takes the standard string format timestamp        *
#*   Used right through the script and parses it and returns         *
#*   an array of different aspects of the content which are          *
#*   used right through the script.                                  *
#*                                                                   *
#*********************************************************************
   #Time stamp comes in as string
    #201602011234
    #which is yyyy mm dd hh mm
    #write-host $theTimeStamp


    if(($TheTimeStamp.length -lt 12) -or ($TheTimeStamp.length -gt 12)) {
        # We have an invalid time stamp which ought to have been stamped on (sorry) before it got here
        # We'll return 1/1/80
        $RetVal = (1980,1,1,1, "00:00", [datetime]::ParseExact("198001010000",'yyyyMMddHHmm',[system.Globalization.CultureInfo]::InvariantCulture), "198001010000")

    } else {

        #what we want is an array containing
        #arr[0] is year (integer)
        #arr[1] is month (integer)
        #arr[2] is week number
        #arr[3] is day (integer)
        #arr[4] is time (string)
        #arr[5] is date (date time field)
        #arr[6] is the original string
        #This probably would be a bit neater as a custom object, but never mind
        try {
            $TheYear = [convert]::ToInt32($TheTimeStamp.Substring(0,4)  )
            $TheMonth = [convert]::ToInt32($TheTimeStamp.Substring(4,2))
            $TheDay = [convert]::ToInt32($TheTimeStamp.Substring(6,2))
            $TheTime = $TheTimeStamp.Substring(8,2) + ":" + $TheTimeStamp.Substring(10,2)
            $TheDate = [datetime]::ParseExact($TheTimeStamp,'yyyyMMddHHmm',[system.Globalization.CultureInfo]::InvariantCulture)
            #the powershell week number isn't properly localised but hey ho, it will do
            $TheWeek = Get-Date $TheDate -UFormat %V
            $RetVal = ($TheYear, $TheMonth, $TheWeek, $TheDay, $TheTime, $TheDate, $TheTimeStamp)
        }

        catch {

            $RetVal = (1980,1,1,1, "00:00", [datetime]::ParseExact("198001010000",'yyyyMMddHHmm',[system.Globalization.CultureInfo]::InvariantCulture), "198001010000")

        }

        #write-host "----------------" -ForegroundColor blue
        #write-host $RetVal[0] -foregroundcolor green
        #write-host $RetVal[1] -foregroundcolor yellow
        #write-host $RetVal[2] -foregroundcolor cyan
        #write-host $RetVal[3] -foregroundcolor white
        #write-host $RetVal[4] -foregroundcolor black
        #write-host $RetVal[6] -foregroundcolor green
        #write-host "----------------" -ForegroundColor blue
    }

    return $RetVal
 }

 
 function GetFTPTimeStampsforDir($DirPath) {
#*********************************************************************
#*                                                                   *
#*   This is the key function, called recursively, which goes        *
#*   through the file system and returns filenames and datestamps    *
#*   for every file in the system from the start point.              *
#*                                                                   *
#*                                                                   *
#*********************************************************************
   # write-host $DirPath -foregroundcolor darkyellow
    try {
   
        # Credentials
        $FTPRequest = [System.Net.FtpWebRequest]::Create("$($ServerURI)$($DirPath)$('/')")
        $FTPRequest.Credentials = New-Object System.Net.NetworkCredential($User,$Pass)
        $FTPRequest.Method = [System.Net.WebRequestMethods+FTP]::ListDirectoryDetails

        # Don't want Binary, Keep Alive unecessary.
        $FTPRequest.UseBinary = $False
        $FTPRequest.KeepAlive = $False

        #Hit the server for the first time for the dir listing
        #in theory the time stamps could be grabbed from this, but the time format isn't great and while it could be hacked
        #I'd want to be very sure it doesn't vary between platforms
        # Yes, you could get file list only, but the detail is used to distinguish files and directories
        #The data typically looks like this

        # drwxrwxrwx    2 30       8            4096 Jan 31  2015 trophies
        # -rwxrwxrwx    1 30       8           18368 Jan 23  2015 europacup-2006.jpg


        $FTPResponse = $FTPRequest.GetResponse()
        $ResponseStream = $FTPResponse.GetResponseStream()
    }

    catch {
        write-host "FTPMon Handled Error -  Failed to get data from FTP Server" -ForegroundColor Red
        write-host "==========================================================" -ForegroundColor Red
        write-host “Exception Type: $($_.Exception.GetType().FullName)” -ForegroundColor Red
        write-host “Exception Message: $($_.Exception.Message)” -ForegroundColor Red
        $ftpfailure = $true
        
    }


    # Create an array of the detailed directory listing
    $StreamReader = New-Object System.IO.Streamreader $ResponseStream
    $DirListing = ""
    $CurLine = ""
    $DirListing = (($StreamReader.ReadToEnd()) -split [Environment]::NewLine)
    $StreamReader.Close()

    # write-host $DirListing -foregroundcolor yellow
    
    # Close the FTP connection so only one is open at a time
    $FTPResponse.Close()
    
    # Loop through the listings. We're going to query each file individually
    foreach ($CurLine in $DirListing) {

        # write-host "Loop Current Line" $CurLine -foregroundcolor green

        # Split line into space separated array
        $LineTok = ($CurLine -split '\ +')

        # Get the filename (Its the 8th and last field so this works even if it contains spaces)
        $CurFile = $LineTok[8..($LineTok.Length-1)]

        # Figure out if it's a FILE. That deals with directories blank lines and other oddities. 
        # if we want to go recursive directory lines will start with d
        $FileFlag = $LineTok[0].StartsWith("-")
        $DirFlag = $LineTok[0].StartsWith("d")

        # Determine what to do next (file or dir?)
        If ($FileFlag) {
            # OK, now we hit the server again with GetDateTimeStamp to get the datestamp in the controlled format
            # much the same as before, but note the ftpwebrequest includes the filename
            # and the method is GetDateTimeStamp
            # at the moment we are not trapping the error for a file that does not exists or deleted since we read it the file list
            # this is left as a future enhancement

            $FTPRequest = [System.Net.FtpWebRequest]::Create("$($ServerURI)$($DirPath)$("/")$($CurFile)")
            $FTPRequest.Credentials = New-Object System.Net.NetworkCredential($User,$Pass)
            $FTPRequest.Method = [System.Net.WebRequestMethods+FTP]::GetDateTimeStamp
            # write-host "File Flag True" $ServerURI $DirPath $CurFile  -foregroundcolor yellow
 
            $response=$FTPRequest.GetResponse()
            # write-host $response -foregroundcolor gray
            # We now have the details, including timestamp of the file that we requested above.  To get the timestamp of the file, we need to look at the property StatusDescription and save it into a temporary variable
            
            $LastMod = $response.LastModified
            $StatusDesc = $response.StatusDescription

            # Don't altogether trust last modified because again that is coming back inpotentially variable human readable format
            # Status Description looks a bit more reliable even if it needs processing. Mind you this varies too.
            # In the example code I plundered for this status code is documented in the format
            # “213 20140713132703.000”
            # on my test system it comes back as 
            # “213 20140713132703”
            # which omits the .000. Anyway we now see that the date is written as YYYYMMDDHHMMSS[.000.]
            # To get the date into a usable format, we need to perform some substring manipulation
            #You’ll note there is a return code of 213 followed by a space and then followed by the date, pobably a newline somewhere as well.  We want to remove the error code at the beginning.  
            #To do that, we can use the LastIndexOf property to find the space character that occurs closest to the right of the string. 
            # We then want to return 1 character after this so we don’t include the space in our result, finally we want the first 12 chars for our time string

            $NewTimeStamp = ExtractTimeStamps($StatusDesc.Substring($StatusDesc.LastIndexOf(” “) + 1,12))

            # Now we'll put together the full current file path
            $FileRecord = $DirPath + "/"+ $CurFile

             #and store them both in the big global arraylist of all the file data
             # note the use of void to hide the return value which otherwise displays on screen intermittent;y (why!?)
            [void]$global:ArrMainList.add(($FileRecord,$NewTimeStamp))

            # finally close the FTP session
            $FTPResponse.Close()

            # write-host "File Flag entry: FileRecord is " $FileRecord "NewTimeStamp is " $NewTimeStamp "Record Ends"

        } elseif ($DirFlag) {
            #write-host "dir Flag Entry" -foregroundcolor yellow
            #if we have a directory flag we want to go all recursive
            #however lets just check that its not a dot file one way or another
            #my test system doesn't return these, but the exxample code suggests that his server does
            $IsDotFile = $CurFile.StartsWith(".")
            if ($IsDotFile) {
                #If we needed to do something clever with .files (might one run into hidden folders starting with .?)
                #then we would do it here, but for the moment the assumption is we can just ignore them
            } else {
                # we recurse into the sub folder
                $ArrFlist = GetFTPTimeStampsforDir $DirPath"/"$CurFile
            }       
        }

    }
    
}



 # Original pseudocode (before learning powershell!)
 # open log file for read
 #    if fail {
 #        alert
 #        create new log file
 #             if fail {
 #                 alert
 #                 halt
 #              } else {
 #                 continue
 #              }
 #     } no else - drops through
 #
 # read log file     
 #     get last_latest_file_time_stamp (may be blank if new file)
 #     get last_total_file_count (may be blank if new file)
 #     get last_log_timestamp       (may be blank if new file)
 #     if fail {
 #         alert
 #         halt
 #      } else {
 #         continue
 #      }
 # close log file
 #
 # open ftp site
 #     if fail {
 #         alert
 #         halt
 #      } else {
 #         continue
 #      }
 #
 # get ftp file list and timestamps
 #     if fail {
 #         alert
 #         halt
 #      } else {
 #         continue
 #      }
 #
 # get current_latest_file_time-stamp
 # calculate current_total_file_count
 # 
 # evaluate file counts and time stamps
 #      if different and not new log file {
 #          alert
 #      }
 # get current_time_stamp
 # evaluate time stamp against log rollover
 #       if rollover necessary {
 #           roll over log files
 #       }       
 #
 # open log file for write
 #     if fail {
 #         alert
 #         halt
 #      } else {
 #         continue
 #      }
 #
 # write log entry
 #
 # close log file
 #
 #end



#*********************************************************************
#*                                                                   *
#*      Here is the Main Code Body start                             *
#*                                                                   *
#*                                                                   *
#*                                                                   *
#*                                                                   *
#*********************************************************************

$SrcUser = $Server + " PSFTPMon <psftpmon@champwilde.f9.co.uk>"
$ServerURI = "ftp://" + $server + "/"
$ConfidenceEmailInterval = $ConfidenceEmailInterval.ToLower()
$Client = $env:COMPUTERNAME


# This is the main array used to collect the data. Making it clobal is the lazy way to preserve it through recursion
# Note we use arraylist because its more tolerant to having lots of additions (and subtractions)
$global:ArrMainList = New-Object System.Collections.ArrayList

# Now we are past the setup stage we put error handling up to max

$ErrorActionPreference = “Stop”

#blank the log record
$logfile=""
$LogReadErrorFlag = $false
#and the mail message
$MsgBody = ""
#and a flag to monitor FTP issues
$ftpfailure = $false



# get current_time_stamp

$CurTimeStamp= ""
$CurTimeStamp = ExtractTimeStamps(Get-Date -UFormat "%Y%m%d%H%M").ToString() 
$CurTime = Get-Date

# open log file and read last line of data

try {
    $logfile = (Get-Content $LogFileLocation)[-1]
    #write-host "Log File contents : " $logfile  -ForegroundColor green
}

catch {
    write-host "FTPMon Handled Error -  Failed to get data from Local Log File" -ForegroundColor Red
    write-host "==============================================================" -ForegroundColor Red
    write-host “Exception Type: $($_.Exception.GetType().FullName)” -ForegroundColor Red
    write-host “Exception Message: $($_.Exception.Message)” -ForegroundColor Red
    #yeah, two flags are unnecessary, but I don't like setting flags in an if statement that depends on them. I think it leads me into confusion
    $logfile = "Absent"
    $LogReadErrorFlag = $true

}

# $logfile should now contain last line of logfile, which will be 12 digit time stamp, numeral (integer), 12 digit time stamp, csv delimited
# or else should be set to absent if the read failed. A Corrupt log file will be unpredictable

if($logfile.length -le 26) {
    #An initial validation if the string length is too short the record won't work and it might break things elsewhere
    $logfile = "Absent"
    $LogReadErrorFlag = $true
}


if ($logfile -match "Absent") {
    if($displayflag) {
            write-host "FTPMon Handled Error -  Log File Failed Validation" -ForegroundColor Red
            write-host "==================================================" -ForegroundColor Red
            write-host "Unexpected Log file validation error - Log File Missing or Invalid." -ForegroundColor Red
    }

} else {
    #write-host "Reading log file " -ForegroundColor green
    #write-host $logfile  -ForegroundColor green


    # read log file     
    #     get last_latest_file_time_stamp (may be blank if new file)
    $Last_Latest_File_Time_Stamp = extracttimestamps($logfile -split  "," )[0].trim()
    #     get last_total_file_count (may be blank if new file)
    #$Last_Total_File_Count = ($logfile -split  "," )[1]

    $Last_Total_File_Count = [convert]::ToInt32(($logfile -split  "," )[1], 10)

    #     get last_log_timestamp       (may be blank if new file)
    $Last_Log_Time_Stamp =  extracttimestamps($logfile  -split  ","  )[2].trim()

    # so we should have got back two 12 digit strings as time stamps
    # and one long integer

    #So we do some simplistic valuation
   #write-host $Last_Latest_File_Time_Stamp[0] -ForegroundColor DarkCyan
   #write-host $Last_Latest_File_Time_Stamp[1] -ForegroundColor blue
   #write-host $Last_Latest_File_Time_Stamp[2] -ForegroundColor green
   #write-host $Last_Latest_File_Time_Stamp[3] -ForegroundColor yellow
   #write-host $Last_Latest_File_Time_Stamp[4] -ForegroundColor darkred
   #write-host $Last_Latest_File_Time_Stamp[5] -ForegroundColor DarkMagenta



   if  (($Last_Latest_File_Time_Stamp[0] -lt 1981) -or ($Last_Latest_File_Time_Stamp[5] -gt $CurTime)) {
        #and it ought to start after 1980 and before now
        #Any problems reading the log records and it will have been set t 1/1/1980
        if($displayflag) {
                write-host "FTPMon Handled Error -  Log File Failed Validation" -ForegroundColor Red
                write-host "==================================================" -ForegroundColor Red
                write-host "Log file validation error - Latest file stamp is not between 1981 and the current date." -ForegroundColor Red
        }
        $LogReadErrorFlag= $true
    }
    if ($Last_Latest_File_Time_Stamp[6].length -ne 12) {
        if($displayflag) {
                write-host "FTPMon Handled Error -  Log File Failed Validation" -ForegroundColor Red
                write-host "==================================================" -ForegroundColor Red
                write-host "Unexpected Log file validation error - Latest Timestamp not 12 chars." -ForegroundColor Red
                #here we're looking at what was stored in the field directly, not a conversion
                #this should be unreachable, was mainly for older versions of the code
        }
        $LogReadErrorFlag= $true
        #it oughtto be 12 chars
    }
    if (($Last_Total_File_Count -match '^\d+$' -eq $false) -or ($Last_Total_File_Count -lt 1)) {
        if($displayflag) {
                write-host "FTPMon Handled Error -  Log File Failed Validation" -ForegroundColor Red
                write-host "==================================================" -ForegroundColor Red
                write-host "Log file validation error - Last file count not numeric or under 1" -ForegroundColor Red
                #OK last file count might be zero if its a new site but why would you be monitoring it. Zero here is almost certainly a problem
        }
        $LogReadErrorFlag= $true
    }
    if ($Last_Log_Time_Stamp[6].length -ne 12) {
        if($displayflag) {
                write-host "FTPMon Handled Error -  Log File Failed Validation" -ForegroundColor Red
                write-host "==================================================" -ForegroundColor Red
                write-host "Unexpected Log file validation error - Last Log time stamp not 12 chars" -ForegroundColor Red
                #this should be unreachable, was mainly for older versions of the code
        }
        $LogReadErrorFlag= $true
    }
    if (($Last_Log_Time_Stamp[0] -lt 2015) -or ($Last_Log_Time_Stamp[5] -gt $CurTime)) {
        if($displayflag) {
                write-host "FTPMon Handled Error -  Log File Failed Validation" -ForegroundColor Red
                write-host "==================================================" -ForegroundColor Red
                write-host "Log file validation error - Last Log time stamp not between 2015 and current date." -ForegroundColor Red
                #Script written in 2016 so a 2015 log date must be invalid!

        }
        $LogReadErrorFlag= $true
    }


}
if ($LogReadErrorFlag) {
    #we have no valid current log file data. Either it was missing or it was corrupt
    #so we set all the values to zero
    if($displayflag) {
            write-host "Log file data is corrupt or absent - setting Last values to Zero" -ForegroundColor Red
            write-host ""
    }
    $Last_Latest_File_Time_Stamp = (1980,1,1,1, "00:00", [datetime]::ParseExact("198001010000",'yyyyMMddHHmm',[system.Globalization.CultureInfo]::InvariantCulture), "198001010000")
    $Last_Total_File_Count = 0
    $Last_Log_Time_Stamp = (1980,1,1,1, "00:00", [datetime]::ParseExact("198001010000",'yyyyMMddHHmm',[system.Globalization.CultureInfo]::InvariantCulture), "198001010000")
}

            
#Now we recursively get the FTP directory data starting in the Base dir
GetFTPTimeStampsforDir $BaseDirectory


#The Global array contains all the time stamps
#we sort it by date (descending) and then file path (ascending)
#so the first line in the array should be the latest file
$global:ArrMainList = $global:ArrMainList | sort-object @{Expression={$_[1]}; Ascending=$false} , @{Expression={$_[0]}; Ascending=$true}
 
$LatestFTPFileCount = $global:ArrMainList.count 
$LatestFTPTimeStamp = $global:ArrMainList[0][1]
$LatestFTPFilePath =  $global:ArrMainList[0][0] 

# an interesting question is what validation should be done on the FTP data.
# at the moment the assumption is that if the program managed to read FTP then what it got back is good, or if what it got back isn't good something
# has gone badly wrong and an alert is no ba thing anyway
 

if($displayflag) {

     write-host "Logfile Data"  -ForegroundColor Green
     write-host "============"   -ForegroundColor Green
     write-host "Last Latest File Timestamp" $Last_Latest_File_Time_Stamp[5]   -ForegroundColor green
     write-host "Last File Count"  $Last_Total_File_Count   -ForegroundColor green
     write-host "Last Log timestamp"  $Last_Log_Time_Stamp[5]   -ForegroundColor Green

     write-host "----------------------------------------------"  -ForegroundColor white

     write-host "Live Data"  -ForegroundColor cyan
     write-host "========="   -ForegroundColor cyan
    #Write-Host "Item A" $global:ArrMainList[0][0] -ForegroundColor cyan
     Write-Host "Latest File Timestamp " $LatestFTPTimeStamp[5] -ForegroundColor cyan
     Write-Host "Latest File Count" $LatestFTPFileCount -ForegroundColor cyan
     Write-Host "Current Timestamp" $CurTimeStamp[5] -ForegroundColor cyan

}

#now we want to do the regular confidence check
 
 
 if($Last_Latest_File_Time_Stamp[6] -match $LatestFTPTimeStamp[6]) {
    #file times match, ALL Good
    $datechange = $false

 } else {
    $datechange = $true
 }


 if($Last_Total_File_Count -eq $LatestFTPFileCount) {
    #file count equal, ALL Good
    $timechange= $false
 } else {
    $timechange= $true
  }



 $MsgSubject = ""
 $MsgBody = ""


if ($ftpfailure) {
    #there was an error on the way reading the ftp data
    $MsgSubject = $MsgSubject + "PSFTPMon FTP Alert for Server" + $server + " "
    $MsgBody  = $MsgBody +   "`nSome or all of the files on " +  $server + " could not be read with FTP by " + $Client
    $MsgBody  = $MsgBody +   "`n==================================================================================`n"  
    $MsgBody  = $MsgBody +   "There was a problem reading the files with FTP on the system " + $server + ". This means that the monitoring script has not checked the server properly. This should be investigated.`n"  

    #Note this is a very long if because if this has failed we want to do next to nothing...

} else {



    if ($LogReadErrorFlag) {
        #we have no valid current log file data. Either it was missing or it was corrupt
        $MsgSubject = $MsgSubject + "PSFTPMon Log Alert for Server" + $server + " "
        $MsgBody  = $MsgBody +   "`nThe Log File for " +  $server + " could not be read on " + $Client
        $MsgBody  = $MsgBody +   "`n====================================================================`n" 
        $MsgBody  = $MsgBody +   "There was a problem reading the log file " + $LogFileLocation + " on the system " + $Client + ". Either the file was missing or the contents were corrupted.`n"  

    } else {



        if ($datechange){
                $MsgSubject = $MsgSubject + "PSFTPMon File Change Alert for Server " + $server + " "

                $MsgBody  = $MsgBody +   "`nFile(s) have been Changed on " + $server 
                $MsgBody  = $MsgBody +   "`n====================================================================`n" 
                $MsgBody  = $MsgBody +  $LatestFTPFilePath + " has a time stamp of " 
                $MsgBody  = $MsgBody +  $LatestFTPTimeStamp[5]  + " and has been altered or uploaded since "  + $Last_Log_Time_Stamp[5] +  ", when the latest filestamp was " 
                $MsgBody  = $MsgBody +  $Last_Latest_File_Time_Stamp[5]  + ".`n`n" 
                $MsgBody  = $MsgBody +   "If there have been authorised changes made to the system then this message is normal, but if there have been no authorised changes then the system should be checked for unauthorised changes.  Other files may also have been changed.`n`n"        

                if($displayflag) {
                    write-host""
                    Write-Host  "File(s) have been Changed on " $server -ForegroundColor red
                    Write-Host  "=================================================================================="  -ForegroundColor red
                    Write-Host  $LatestFTPFilePath " has a time stamp of "  -ForegroundColor red
                    Write-Host $LatestFTPTimeStamp[5] " and has been altered or uploaded since " $Last_Log_Time_Stamp[5] "," "when the latest filestamp was " -ForegroundColor red
                    write-host $Last_Latest_File_Time_Stamp[5] ". Other files may also have been changed." -ForegroundColor red
                }    
            }

        if ($timechange) {
            $MsgSubject = $MsgSubject + "PSFTPMon File Addition/Removal Alert for Server " + $server + " "

            $MsgBody  = $MsgBody +   "`nFile(s) have been added or removed on "  +  $server 
            $MsgBody  = $MsgBody +   "`n====================================================================`n"  
            $MsgBody  = $MsgBody +   "There are now " +  $LatestFTPFileCount +  " files, but there were " +  $Last_Total_File_Count  + " on "  + $Last_Log_Time_Stamp[5]  + ".`n`n"  
            $MsgBody  = $MsgBody +   "If there have been authorised changes made to the system then this message is normal, but if there have been no authorised changes then the system should be checked for unauthorised changes.`n`n"        

             if($displayflag) {

                write-host""
                Write-Host  "File(s) have been added or removed on " $server -ForegroundColor red
                Write-Host  "=================================================================================="  -ForegroundColor red
                Write-Host  "There are now " $LatestFTPFileCount " files, but there were " $Last_Total_File_Count " on " $Last_Log_Time_Stamp[5] "."  -ForegroundColor red
             }
        }
    }


     # evaluate time stamp against log rollover
     #       if rollover necessary {
     #           roll over log files
     #       }       
     #
 
 
     # open log file for write
     #     if fail {
     #         alert
     #         halt
     #      } else {
     #         continue
     #      }
     #


    $Rotate_Log_Conf_Mail = $false

    Switch ($ConfidenceEmailInterval) {
        #day changes when [3] changes
        #week changes when [2] changes
        #month changes when [1] changes
        #year changes when [0] changes
        "daily"    { if ($CurTimeStamp[3] -ne  $Last_Log_Time_Stamp[3]){$Rotate_Log_Conf_Mail = $true}}
        "weekly"   { if ($CurTimeStamp[2] -ne  $Last_Log_Time_Stamp[2]){$Rotate_Log_Conf_Mail = $true}}
        "monthly"  { if ($CurTimeStamp[1] -ne  $Last_Log_Time_Stamp[1]){$Rotate_Log_Conf_Mail = $true}}
        "annual"   { if ($CurTimeStamp[0] -ne  $Last_Log_Time_Stamp[0]){$Rotate_Log_Conf_Mail = $true}}
        default {}
    }

    If ($Rotate_Log_Conf_Mail) {
        if ($LogReadErrorFlag -ne $true) {

                $MsgSubject = $MsgSubject + "PSFTPMon " + $ConfidenceEmailInterval + " Monitoring Message for Server " + $server + " "

                $MsgBody  = $MsgBody +   "`nThis is the " + $ConfidenceEmailInterval + " monitoring message for server " +  $server 
                $MsgBody  = $MsgBody +   "`n====================================================================`n" 
                $MsgBody  = $MsgBody +   "This message is expected and does not require action unless there are other alerts above this text. "
                $MsgBody  = $MsgBody +   "It is sent simply to inform you that the system " + $client
                $MsgBody  = $MsgBody +   " is still monitoring server " + $server + " for changes.`n`n"
                $MsgBody  = $MsgBody +  " The last recorded change was to " + $LatestFTPFilePath + " on " +  $LatestFTPTimeStamp[5] + ".`n`n"

                #finally we rotate the log files
 
                #set error action to silently continue since its not the end of the world if log files get messed up.
                $ErrorActionPreference = “SilentlyContinue”
                #$LogFileLocation = "C:\logs\DonAshton.log"
                #this is the number of backup log files to keep
                #$NoBackupLogs = 5

                $lpath =  $LogFileLocation.Substring(0,$LogFileLocation.LastIndexOf(”\“) + 1)
                $fpath =  $LogFileLocation.Substring($LogFileLocation.LastIndexOf(”\“) + 1)
                $fname =  $fpath.Substring(0,$fpath.LastIndexOf(”.“) + 1)

                $OldestLog = ($lpath + $fname + "bk" + $NoBackupLogs).Trim()
                remove-item $OldestLog

                for($i=$NoBackupLogs; $i -gt 1; $i--){
                    $NameFrom = ($lpath + $fname + "bk" + ($i - 1)).Trim()
                    $NameTo = ( $fname + "bk" + $i).Trim()
                    Rename-Item $NameFrom $NameTo
                }
                $NameTo = ( $fname + "bk1").Trim()
                Rename-Item $LogFileLocation $NameTo

                $ErrorActionPreference = “Stop”
        }
    }

     # write log entry. Note we don't write a log file if we got a bad ftp read, but we do write an entry if we got a bad log file
     # apart from anything else that gives us a new log file

    try {
        if($Rotate_Log_Conf_Mail){
            #there's an interesting feature in powershell, at least on what I'm running.
            #if get-content is run against a single file with the -1 parameter then it gets the
            #last character of the file, not the last line
            #hence here we detect if we've just rolled over a log file, and if so create a header line so 
            #the new file starts off with two lines. [FX: sigh]
            $logline = "LatestFTPTimeStamp, LatestFTPFileCount, CurTimeStamp"
            out-file -filepath $LogFileLocation -inputobject $logline -encoding ASCII -Append -NoClobber
        }
        $logline = $LatestFTPTimeStamp[6] + "," + $LatestFTPFileCount + "," + $CurTimeStamp[6]
        out-file -filepath $LogFileLocation -inputobject $logline -encoding ASCII -Append -NoClobber
     }
    catch {
        write-host "FTPMon Handled Error -  Failed to write to Local Log File" -ForegroundColor Red
        write-host "=========================================================" -ForegroundColor Red
        write-host “Exception Type: $($_.Exception.GetType().FullName)” -ForegroundColor Red
        write-host “Exception Message: $($_.Exception.Message)” -ForegroundColor Red
        $MsgSubject = $MsgSubject + "PSFTPMon Log Alert for Server" + $server + " "
        $MsgBody  = $MsgBody +   "`nThe Log File for " +  $server + " could not be written to on " + $Client
        $MsgBody  = $MsgBody +   "`n==================================================================================`n"  
        $MsgBody  = $MsgBody +   "There was a problem writing the log file " + $LogFileLocation + " on the system " + $Client + ". The file may have been locked or could not be created.`n"  
    }



}
 
if($MsgBody.length -gt 0) {
   $MsgBody  = $MsgBody +   "`nThis is an automated message from system " + $client + ". Please do not reply to it.`n"

   send-mailmessage -to $DestUsers -from $SrcUser -subject $MsgSubject -body  $MsgBody -SmtpServer $smtpserver  
}

 


<#
Testing the Script
==================
This one is complex enough that you really should test it before running live, especially as its kinda beginner coding at the moment
This is probably a reasonable starting point

FTP Testing
===========
1) Server unavailable.                                                     
Test - if using a test sstem take it down!
3) File Added
Create a file on the server. Check it alerts
3) File modified - change the file you just created,check it alerts
4) File Deleted - remove the file you created, check it alerts
5) No change - run serveral times, check it doesn't alert.


Log File Testing
================
1) File missing                                                             
Delete log file and see if it alerts and recreates, and then runs OK
2) File corrupt
delete characters from last line of log file, see what happens             
add random characters to last line of log file, see what happens

General Function
================
1) Log Rollover (Change Last Log date, year month etc)                  
1.1) Check Daily, weekly, monthly etc
2) Change last file update, correct warnings?                            
3) Change File count correct warnings 
3.1) One less                                                           
3.2) One more                                                         
4) Normal Operation                                                      
5) NormalOperation no display                                           
6) Simultaneous count and timetsamp change                            
7) Simultaneous count and timestamp change plus rollover                                





#>




<#

Powershell Scripts should be digitally signed on your system so that they can be run without crippling
some of the system security. Its a complex topic, but for running scripts on a single machine this guide
seems to work well for me.

http://www.hanselman.com/blog/SigningPowerShellScripts.aspx

#>